"""
Receive warning and error messages from the ESP8266 board.
Look up which email/SMS belongs to which board.
Send email/SMS to user whose board is reporting an error.

Google Form can be found at https://forms.gle/fwnfND1UR8HPRLcd7
The spreadsheets can be found in https://drive.google.com/drive/folders/1ljhUU-meD42erOlxsccloBpux__nhh7Q
"""

from __future__ import print_function

import csv
import os
import signal
import smtplib
import ssl
import sys
from email.message import EmailMessage

import paho.mqtt.client as mqtt
import pendulum
from apscheduler.schedulers.background import BackgroundScheduler
from google.auth.exceptions import RefreshError
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build
from googleapiclient.errors import HttpError
from requests import get
from twilio.rest import Client

DEBUG_MODE = True  # Outputs additional console messages when true.
SEND_MODE = True  # Sets whether the code should send email/text alerts
KEEPALIVE_PERIOD = 45  # Roughly how often the board will send a keepalive message, in seconds.
KEEPALIVE_CUTOFF = (KEEPALIVE_PERIOD * 2.01) + 10  # How long since last keepalive to consider the board unresponsive.
KEEPALIVECHECK_PERIOD = (KEEPALIVE_PERIOD / 24) + 45  # How often the code will check for dead boards.
KEEPALIVE_FILE = 'keepalives.csv'  # The file that will store the keepalive history of all boards.
BADDEVICES_FILE = 'baddevices.csv'  # The file that will store the unresponsive devices.
LOGFILE = 'pipefreeze.log'  # The file which will log all console and debug messages

# ID and range of the Google Spreadsheet which contains the data directly outputted by the Google Form.
FORM_RESPONSES_SHEET_ID = '1WvVXJM0RNQN10BOLY0aynwz_zVXWyLuvTQCeJAjiTf4'
RAW_STUDENT_DATA_SHEET_RANGE = 'A2:F'

# ID and ranges of the Google Spreadsheet which contains all other relevant information.
MAIN_SHEET_ID = '1FBZFUN2xhx0kOCmUmhXGizsMhFGUdgip6Jxez7v_7pY'
DEVICE_REGISTRY_SHEET_RANGE = 'Device Registry!A2:C'
DORM_CONTACT_SHEET_RANGE = 'Dorm Contacts!A2:D'
MAINTENANCE_CONTACT_SHEET_RANGE = 'Maintenance Contact!A2'

# MQTT server configuration
MQTT_SERVER_LOCAL_ADDRESS = 'localhost'  # IP of the local computer, use when code is run on the Raspberry Pi
MQTT_SERVER_GLOBAL_ADDRESS = '73.50.54.233'  # Global IP of the network the server is located on
MQTT_SERVER_UNAME = 'pipefreeze'  # Username of the MQTT server, defined on the Raspberry Pi
MQTT_SERVER_PASSWORD = 'almamater'  # Password of the MQTT server, defined on the Raspberry Pi
MQTT_SERVER_TOPIC = 'final'  # Name of the topic to subscribe to. Must match the value that the ESP is using.

# Gmail account created specifically for the project
EMAIL_SENDER_ADDRESS = 'pipefreeze1842@gmail.com'  # Email address that all communications will come from
EMAIL_SENDER_PASSWORD = '@lmamater'  # Google account password

# Found at https://console.twilio.com/
TWILIO_ACCOUNT_SID = 'AC01370d9889517eedaeb4978d857a1114'
TWILIO_AUTH_TOKEN = '4161ed9f720db1a141daac1c30047ec9'
TWILIO_SENDER_PHONENUM = '+18053869577'  # Phone number that all text messages will come from

# Time zone configuration settings (using Pendulum's time string formatting)
GSHEET_TIME_FORMAT = 'M/D/YYYY H:mm:ss'  # Format that the form responses are recorded in
CONSOLE_TIME_FORMAT = 'YYYY-MM-DD HH:mm:ss.SSS zz'  # Format to be outputted with the console messages
CSV_TIME_FORMAT = 'YYYY-MM-DD HH:mm:ss.SSS'  # Format to be outputted to text files
TIME_ZONE = 'America/Indiana/Indianapolis'  # Time zone for the form responses


def sigint_handler(_sig, _frame):
    output_debug_message('Closing program cleanly.')
    global mqttclient
    mqttclient.disconnect()
    scheduler.remove_all_jobs()
    scheduler.shutdown()
    sys.exit()


def generate_console_message(message):
    current_time = pendulum.now().format(CONSOLE_TIME_FORMAT)
    consolemessage = current_time + '   |   ' + f'{message}'
    return consolemessage


def output_console_message(message):
    # Gets the current time and prints a timestamped message in the console
    console_message = generate_console_message(message)
    print(console_message)
    with open(LOGFILE, mode='a') as f:
        f.write(console_message + '\n')


def output_debug_message(debug_message):
    # Prints message to the console, but only if debug mode is activated
    if DEBUG_MODE:
        output_console_message(debug_message)
    else:
        console_message = generate_console_message(debug_message)
        with open(LOGFILE, mode='a') as f:
            f.write(console_message + '\n')


def transpose_matrix(matrix):
    # Transpose the rows and columns of a "matrix" (a.k.a. a two-dimensional list object)
    matrix_t = [list(x) for x in zip(*matrix)]
    return matrix_t


def keepalive(macaddress):
    # Record the keepalive message from the board to a file.
    # File is a CSV file with format MACADDRESS,TIMESTAMP\r\n

    # Check if the keepalive file exists.
    try:
        with open(KEEPALIVE_FILE, mode='x', newline=''):
            output_debug_message('Keepalive record file does not exist. Creating new one.')
    except FileExistsError:
        output_debug_message('Keepalive record file found. Reading file.')

    # Check if the device that sent the keepalive has sent one before.
    with open(KEEPALIVE_FILE, mode='r', newline='') as f:
        reader = csv.reader(f)  # Read the file
        readerlist = list(reader)  # Convert to a list

    if any(macaddress in line for line in readerlist):
        output_debug_message('Keepalive record found for this device. Updating entry.')
        # Update the timestamp for the given board.
        with open(KEEPALIVE_FILE, mode='w', newline='') as f:
            writer = csv.writer(f)
            for line in readerlist:
                if line[0] == macaddress:
                    line[1] = pendulum.now().format(CSV_TIME_FORMAT)
            writer.writerows(readerlist)
    else:
        output_debug_message('No keepalive record found for this device. Creating entry.')
        # Create a new entry for the given board at the end of the file.
        with open(KEEPALIVE_FILE, mode='a', newline='') as f:
            writer = csv.writer(f)
            writer.writerow([macaddress, pendulum.now().format(CSV_TIME_FORMAT)])

    if os.path.exists(BADDEVICES_FILE):
        with open(BADDEVICES_FILE, mode='r', newline='') as f:
            baddevicelist = list(csv.reader(f))
        if any(macaddress in line for line in baddevicelist):
            output_debug_message(f'Removing device {macaddress} from unresponsive device list.')
            with open(BADDEVICES_FILE, mode='w', newline='') as f:
                writer = csv.writer(f)
                baddevicelist.remove([macaddress])
                writer.writerows(baddevicelist)


def check_keepalives():
    keepalive_cutoff = pendulum.duration(seconds=KEEPALIVE_CUTOFF)
    output_debug_message('Checking for unresponsive boards.')

    if os.path.exists(KEEPALIVE_FILE):
        with open(KEEPALIVE_FILE, mode='r', newline='') as f:
            keepalivereader = csv.reader(f)  # Read the file
            keepalivelist = list(keepalivereader)  # Convert to a list

        if os.path.exists(BADDEVICES_FILE):
            with open(BADDEVICES_FILE, mode='r', newline='') as f:
                badreader = csv.reader(f)
                badlist = list(badreader)
        else:
            badlist = []

        for line in keepalivelist:
            last_keepalive = pendulum.from_format(line[1], CSV_TIME_FORMAT, tz=TIME_ZONE)
            unresponsive_time = pendulum.now() - last_keepalive
            if unresponsive_time > keepalive_cutoff:
                macaddress = line[0]
                if not any(macaddress in line for line in badlist):
                    with open(BADDEVICES_FILE, mode='a', newline='') as f:
                        csv.writer(f).writerow([macaddress])
                    output_debug_message(f'Device {macaddress} has not responded in '
                                         f'{unresponsive_time.in_words()}. '
                                         f'Sending alerts.')
                    send_alerts(macaddress, -3, 'NaN')


def read_gsheet(spreadsheet_id, range_name):
    # Gets the data from the given Google spreadsheet.

    scopes = ['https://www.googleapis.com/auth/spreadsheets']

    def make_creds():
        output_debug_message('Rebuilding Google credentials.')
        flow = InstalledAppFlow.from_client_secrets_file('credentials.json', scopes)
        newcreds = flow.run_local_server(port=0)
        return newcreds

    def save_creds(working_creds):
        output_debug_message('Saving Google credentials for later use.')
        with open('token.json', 'w') as newtoken:
            newtoken.write(working_creds.to_json())

    def reset_creds():
        output_debug_message('Deleting and resetting credentials.')
        os.remove('token.json')
        goodcreds = make_creds()
        save_creds(goodcreds)
        return goodcreds

    if os.path.exists('token.json'):
        creds = Credentials.from_authorized_user_file('token.json', scopes)
        if not creds.valid:
            creds.refresh(Request())
            if not creds.valid:
                creds = reset_creds()
    else:
        output_debug_message('Token file not found. Making new credentials.')
        creds = make_creds()
        save_creds(creds)

    try:
        service = build('sheets', 'v4', credentials=creds)
        sheet = service.spreadsheets()
        sheetdata = sheet.values().get(spreadsheetId=spreadsheet_id,
                                       range=range_name,
                                       majorDimension='COLUMNS').execute()
        return sheetdata['values']
    except RefreshError as err:
        output_debug_message(f'A refresh error occurred: {err}')
        _creds = reset_creds()
        sys.exit()
    except HttpError as err:
        output_debug_message(f'An HTTP error occurred: {err}')


def filter_student_data(raw_student_data):
    student_data = raw_student_data.copy()  # Create a copy of the student spreadsheet data to be modified

    # Convert the Google form timestamps to Pendulum datetime objects
    student_data[0] = [pendulum.from_format(x, GSHEET_TIME_FORMAT, tz=TIME_ZONE) for x in student_data[0]]

    # Each entry in the matrix contains all the answers to only one question.
    # In other words, matrix[0] contains all the timestamps, matrix[1] contains all the dorm names, etc.
    # This must be changed, so that matrix[0] contains all the data from the first person to fill out the form.
    # Transpose the rows and columns, so each entry in the matrix contains one form response.
    student_data_transposed = transpose_matrix(student_data)

    # Sort the matrix in chronological order.
    timesorted_student_data = sorted(student_data_transposed, key=lambda x: x[0])
    # Create a new data point for each entry, containing a concatenated string of the dorm name and room number.
    timesorted_student_data = [x + [x[1] + ' ' + x[2]] for x in timesorted_student_data]

    filtered_student_data = []
    preexisting_dormrooms = []
    for x in timesorted_student_data[::-1]:  # iterate through the list backwards (reverse chronological order)
        if x[6] not in preexisting_dormrooms:  # x[6] contains the concatenated string of the dorm name and room number
            preexisting_dormrooms.append(x[6])  # no duplicates allowed
            filtered_student_data.append(x)

    filtered_student_data = transpose_matrix(filtered_student_data)
    return filtered_student_data


def sendemail(receiver_email, emailsubject, emailmessage):
    # Send the given email message to the given recipient.
    if SEND_MODE:
        port = 465
        smtp_server = 'smtp.gmail.com'
        sender_email = EMAIL_SENDER_ADDRESS
        password = EMAIL_SENDER_PASSWORD

        msg = EmailMessage()  # Creating a new email message object to hold all the data
        msg['from'] = sender_email
        msg['to'] = receiver_email
        msg['subject'] = emailsubject
        msg.set_content(emailmessage)

        output_console_message(f'Sending email message to {receiver_email}')
        context = ssl.create_default_context()
        with smtplib.SMTP_SSL(smtp_server, port, context=context) as emailserver:
            emailserver.login(sender_email, password)
            emailserver.send_message(msg)


def sendsms(receiver_sms, smsmessage):
    # Send the given SMS message to the given recipient, using the Twilio SMS API.

    if SEND_MODE:
        account_sid = TWILIO_ACCOUNT_SID
        auth_token = TWILIO_AUTH_TOKEN
        sender_sms = TWILIO_SENDER_PHONENUM
        client = Client(account_sid, auth_token)

        output_console_message(f'Sending SMS message to {receiver_sms}')
        client.messages.create(
            body=smsmessage,
            from_=sender_sms,
            to=receiver_sms
        )


def send_alerts(mac_address, error_code, temperature):
    student_data = filter_student_data(read_gsheet(FORM_RESPONSES_SHEET_ID, RAW_STUDENT_DATA_SHEET_RANGE))
    device_data = read_gsheet(MAIN_SHEET_ID, DEVICE_REGISTRY_SHEET_RANGE)
    dorm_data = read_gsheet(MAIN_SHEET_ID, DORM_CONTACT_SHEET_RANGE)
    maintenance_data = read_gsheet(MAIN_SHEET_ID, MAINTENANCE_CONTACT_SHEET_RANGE)

    # Find the dorm name and room number associated with the given MAC address.
    try:
        registry_index = device_data[0].index(mac_address)
    except ValueError:
        output_console_message('Error: The device which just sent a message is not found in the list of devices. '
                               'Ignoring.')
        return
    device_dorm = device_data[1][registry_index]
    device_room = device_data[2][registry_index]
    device_dormroom = device_dorm + ' ' + device_room  # Create a concatenated dormroom string, for comparison

    # Find the student name, email, and phone number that corresponds with the given dorm room.
    try:
        student_contact_index = student_data[6].index(device_dormroom)
    except ValueError:
        output_console_message('Error: No student is registered to the given dormroom. Ignoring.')
        return
    student_name = student_data[3][student_contact_index]
    student_email = student_data[4][student_contact_index]
    student_phonenum = '+1' + student_data[5][student_contact_index]

    # Find the dorm contact's email and phone number that corresponds with the given dorm name.
    try:
        dormcontact_index = dorm_data[0].index(device_dorm)
    except ValueError:
        output_console_message('Error: No dorm contact is on file for the given dorm. Ignoring.')
        return
    dormcontact_email = dorm_data[2][dormcontact_index]
    dormcontact_phonenum = '+1' + dorm_data[3][dormcontact_index]

    # Get the email address for the maintenance department.
    maintenance_email = maintenance_data[0][0]

    match error_code:
        case 1:
            # Warning alert (Temperature <= 10°C)
            alert_content = f'WARNING: Your radiator is at an unsafe temperature ({temperature}°C). ' \
                            f'Please close your window to prevent serious damage.'
            alert_subject = 'RADIATOR TEMPERATURE WARNING'
            sendemail(student_email, alert_subject, alert_content)
            sendsms(student_phonenum, alert_content)

        case 2:
            # Critical alert (Temperature <= 5°C)
            student_alert_content = f'EMERGENCY: Your radiator is at a dangerous temperature ({temperature}°C). ' \
                                    f'PLEASE close your window IMMEDIATELY to prevent serious damage.'
            student_alert_subject = 'RADIATOR TEMPERATURE EMERGENCY'
            sendemail(student_email, student_alert_subject, student_alert_content)
            sendsms(student_phonenum, student_alert_content)

            dormcontact_alert_content = f'EMERGENCY: {student_name}’s radiator in {device_dorm} room {device_room} ' \
                                        f'is at a critical temperature. Serious damage is imminent.'
            dormcontact_alert_subject = f'RADIATOR TEMPERATURE EMERGENCY IN ROOM {device_room}'
            sendemail(dormcontact_email, dormcontact_alert_subject, dormcontact_alert_content)
            sendsms(dormcontact_phonenum, dormcontact_alert_content)

        case -1:
            # Low battery alert
            dormcontact_alert_content = f'ALERT: The pipe freeze detector in {device_dorm} room {device_room} ' \
                                        f'is out of battery. Please arrange to have the battery replaced.'
            dormcontact_alert_subject = f'Pipe Freeze Detector – Dead Battery in Room {device_room}'
            sendemail(dormcontact_email, dormcontact_alert_subject, dormcontact_alert_content)
            sendsms(dormcontact_phonenum, dormcontact_alert_content)

            maintenance_alert_content = f'ALERT: The pipe freeze detector in {device_dorm} room {device_room} ' \
                                        f'is out of battery. Please arrange to have the battery replaced.'
            maintenance_alert_subject = f'Pipe Freeze Detector – Dead Battery in {device_dorm} Room {device_room}'
            sendemail(maintenance_email, maintenance_alert_subject, maintenance_alert_content)

        case -2:
            # Sensor error alert
            dormcontact_alert_content = f'ALERT: The temperature sensors in {device_dorm} room {device_room} ' \
                                        f'are malfunctioning. Please refer the unit to maintenance.'
            dormcontact_alert_subject = f'Pipe Freeze Detector – Temperature Sensor Malfunction ' \
                                        f'in Room {device_room}'
            sendemail(dormcontact_email, dormcontact_alert_subject, dormcontact_alert_content)
            sendsms(dormcontact_phonenum, dormcontact_alert_content)

            maintenance_alert_content = f'ALERT: The temperature sensors in {device_dorm} room {device_room} ' \
                                        f'are malfunctioning. Please arrange to have the device serviced.'
            maintenance_alert_subject = f'Pipe Freeze Detector – Temperature Sensor Malfunction ' \
                                        f'in {device_dorm} Room {device_room}'
            sendemail(maintenance_email, maintenance_alert_subject, maintenance_alert_content)

        case -3:
            # No recent keepalive message received alert
            dormcontact_alert_content = f'ALERT: The pipe freeze detector in {device_dorm} room {device_room} ' \
                                        f'has stopped checking in with the server. There may be a connection error. ' \
                                        f'Please refer the unit to maintenance.'
            dormcontact_alert_subject = f'Pipe Freeze Detector – Unresponsive Board ' \
                                        f'in Room {device_room}'
            sendemail(dormcontact_email, dormcontact_alert_subject, dormcontact_alert_content)
            sendsms(dormcontact_phonenum, dormcontact_alert_content)

            maintenance_alert_content = f'ALERT: The pipe freeze detector in {device_dorm} room {device_room} ' \
                                        f'has stopped checking in with the server. There may be a connection error. ' \
                                        f'Please arrange to have the device serviced.'
            maintenance_alert_subject = f'Pipe Freeze Detector – Unresponsive Board ' \
                                        f'in {device_dorm} Room {device_room}'
            sendemail(maintenance_email, maintenance_alert_subject, maintenance_alert_content)


def read_mqtt(serverip, uname, pword, topic):
    def on_connect(mqtt_client, _userdata, _flags, rc):  # The callback for when the client connects to the broker
        output_debug_message(f'Connected with result code {rc}')
        output_debug_message(f'Now subscribing to topic \'{topic}\'')
        mqtt_client.subscribe(topic)  # Subscribe to the given topic, receive any messages published on it

    def on_message(_mqtt_client, _userdata, msg):  # The callback for when a PUBLISH message is received.
        msgpayload = msg.payload.decode('utf-8')
        output_debug_message(f'Message received: \'{msgpayload}\'')  # Print a received message

        # Format for messages: 0f:0f:0f:0f:0f:0f_0_-12
        #        a.k.a.        MACADDRESS_ERRORCODE_TEMPERATURE
        msgsplitup = msgpayload.split('_')
        msg_macaddress = msgsplitup[0]  # MAC address of the device that sent the message
        try:
            msg_temperature = float(msgsplitup[2])  # The temperature reading which is causing the alert (°C)
        except ValueError:
            output_console_message('Error: Invalid temperature received. Ignoring.')
            return

        # Error codes:
        # 2 = emergency
        # 1 = warning
        # 0 = device sending keepalive message
        # -1 = low battery
        # -2 = sensor malfunction
        # -3 = device has not sent keepalive lately

        # If the error code is 0, -1, -2, or -3, the temperature will be reported as 'NaN'.

        msg_errorcode = int(msgsplitup[1])

        match msg_errorcode:
            case (1 | 2 | -1 | -2):
                send_alerts(msg_macaddress, msg_errorcode, msg_temperature)
            case 0:
                keepalive(msg_macaddress)
            case _:
                output_console_message(f'Message has an invalid error code: {msg_errorcode}. Ignoring.')

    global mqttclient
    mqttclient = mqtt.Client()  # Create instance of client with random Client ID
    mqttclient.on_connect = on_connect  # Define callback function for successful connection
    mqttclient.on_message = on_message  # Define callback function for receipt of a message
    mqttclient.username_pw_set(uname, password=pword)
    mqttclient.connect(serverip)  # Connect to the given MQTT server
    mqttclient.loop_forever()  # Start networking daemon


if __name__ == '__main__':
    globalip = get('https://api.ipify.org').content.decode('utf8')

    if globalip == MQTT_SERVER_GLOBAL_ADDRESS:
        MQTT_SERVER_ADDRESS = MQTT_SERVER_LOCAL_ADDRESS
        output_debug_message(f'Local server detected. Running with host IP \'{MQTT_SERVER_ADDRESS}\'')
    else:
        MQTT_SERVER_ADDRESS = MQTT_SERVER_GLOBAL_ADDRESS
        output_debug_message(f'Global server detected. Running with host IP \'{MQTT_SERVER_ADDRESS}\'')

    scheduler = BackgroundScheduler()
    keepalivejob = scheduler.add_job(check_keepalives, 'interval', seconds=KEEPALIVECHECK_PERIOD)
    scheduler.start()

    signal.signal(signal.SIGINT, sigint_handler)

    output_console_message('Connecting to MQTT server.')
    global mqttclient
    read_mqtt(MQTT_SERVER_ADDRESS, MQTT_SERVER_UNAME, MQTT_SERVER_PASSWORD, MQTT_SERVER_TOPIC)
